(************************************************************************)
(*                                                                      *)
(*    MACHINE LEARNING PLUGIN FOR COQ                                   *)
(*    Jonathan Laurent, ENS Paris - 2016                                *)
(*                                                                      *)
(************************************************************************)


DECLARE PLUGIN "plugin"

open Globnames
open Pp
open Printf

let debug_mode = ref false

let colt_dir = ref ""

let model_file = ref "model"

(************************************************************************)

(** {6 Utilities} *)

let (||>) f g x = x |> f |> g

let read_lines chan = 
  let lines = ref [] in
  try
    while true; do
      lines := input_line chan :: !lines
    done; !lines
  with End_of_file ->
    close_in chan;
    List.rev !lines


(************************************************************************)


(** {6 Features definitions} *)

let string_of_gref gref = 
  Libnames.string_of_path (Nametab.path_of_global gref)


module Feature = struct

  type t = 
  | Dep of global_reference
  | Ref of global_reference

  let to_string = function
    | Dep r -> sprintf "dep:%s" (string_of_gref r)
    | Ref r -> sprintf "ref:%s" (string_of_gref r)

end


module Features = struct

  type t = Feature.t list

  let empty = []

  let add f fs = f :: fs

  let concat = List.concat

  let filter = List.filter

  let to_list fs = List.sort_uniq compare fs

  let to_string ?(sep="\n") fs = 
    fs |> List.map Feature.to_string 
       |> String.concat sep

end

(************************************************************************)

(** {6 Features search} *)

exception Not_closed

let traverse_term f (c:Term.constr) acc =
  let rec tr c acc =
    match Term.kind_of_term c with
      Term.Rel _ -> acc
    | Term.Var x -> acc
    | Term.Sort s -> acc
    | Term.Meta _ -> raise Not_closed
    | Term.Evar _ -> raise Not_closed
    | Term.Cast(c,_,t) -> tr c (tr t acc)
    | Term.Prod(n,t,c) -> tr t (tr c acc)
    | Term.Lambda(n,t,c) -> tr t (tr c acc)
    | Term.LetIn(_,v,t,c) -> tr v (tr t (tr c acc))
    | Term.App(c,ca) -> tr c (Array.fold_right tr ca acc)
    | Term.Const cst -> 
      f (Globnames.ConstRef (Univ.out_punivs cst)) acc
    | Term.Ind i -> f (Globnames.IndRef (Univ.out_punivs i)) acc
    | Term.Construct cnst -> 
      f (Globnames.ConstructRef (Univ.out_punivs cnst)) acc
    | Term.Case({Term.ci_ind=i},c,t,ca) ->
      f (Globnames.IndRef i) 
        (tr c (tr t (Array.fold_right tr ca acc)))
    | Term.Fix(_,(_,ca,ca')) -> 
      Array.fold_right tr ca (Array.fold_right tr ca' acc)
    | Term.CoFix(_,(_,ca,ca')) -> 
      Array.fold_right tr ca (Array.fold_right tr ca' acc)
    | Term.Proj(p, c) ->
      tr c acc
  in tr c acc


(* Adapted from 'coq-dpdgraph' *)

let is_prop_cache = Hashtbl.create 100000

let is_prop gref =
  let s = string_of_gref gref in
  try Hashtbl.find is_prop_cache s
  with Not_found ->
    let b = 
      try
        let glob = Glob_term.GRef(Loc.ghost, gref, None) in
        let env = Global.env() in
        let sigma = Evd.from_env env in
        let sigma', c = Pretyping.understand_tcc env sigma glob in
        let sigma2 = Evarconv.consider_remaining_unif_problems env sigma' in
        let sigma3, nf = Evarutil.nf_evars_and_universes sigma2 in
        let pl, uctx = Evd.universe_context sigma3 in
        let env2 = Environ.push_context uctx (Evarutil.nf_env_evar sigma3 env) in
        let c2 = nf c in
        let t = Environ.j_type (Typeops.infer env2 c2) in
        let t2 = Environ.j_type (Typeops.infer env2 t) in
        Term.is_Prop t2
      with _ -> false in
    Hashtbl.add is_prop_cache s b ;
    b


exception Error of string

let collect_features t = 
  let add_ref r = Features.add (Feature.Ref r) in
  traverse_term add_ref t Features.empty


let features gref = 

  let add_dep d acc = 
    if is_prop d then
      Features.add (Feature.Dep d) acc
    else 
      acc in

  match gref with
  | VarRef _ | IndRef _ -> raise (Error "Not supported.")
  | ConstRef cst ->

    let cb = Environ.lookup_constant cst (Global.env()) in
    
    (* Check dependencies *)
    let deps = 
      match Global.body_of_constant_body cb with
      | Some t -> traverse_term add_dep t Features.empty
	  | None -> Features.empty in

    (* References to constants *)
    let refs = 
      match  cb.Declarations.const_type with
        | Declarations.RegularArity t -> collect_features t
        | Declarations.TemplateArity _ ->  Features.empty in

    Features.concat [deps; refs]



let print_features gref =
  try 
    let fs = features gref in
    Features.to_list fs |> List.iter (Feature.to_string ||> str ||> Pp.msgnl)
  with Error s -> Pp.msgerrnl (str s)


let write_features gref filename = 
  try
    let fs = features gref in
    (* let oc = open_out_gen [ Open_append; Open_creat; Open_text] 0o777 filename in *)
    let oc = open_out filename in
    Features.to_list fs |> List.iter (fun f ->
      output_string oc (Feature.to_string f);
      output_char oc '\n'
    );
    close_out oc
  with Error _ -> ()


(************************************************************************)

(** {6 Vernacular commands} *)


VERNAC COMMAND EXTEND Features
  | ["Features" reference(ref)] -> [ print_features (Nametab.global ref) ]
END

VERNAC COMMAND EXTEND WriteFeatures
  | ["Write" "Features" reference(ref) "in" string(f) ] -> 
    [ write_features (Nametab.global ref) f ]
END

VERNAC COMMAND EXTEND DebugMode
  | ["Colt" "Set" "DebugMode"] -> 
    [ debug_mode := true ]
END

VERNAC COMMAND EXTEND SetColtDir
  | ["Colt" "Set" "LoadPath" string(ref)] -> 
    [ colt_dir := ref ]
END

VERNAC COMMAND EXTEND SetModel
  | ["Colt" "Set" "Model" string(ref)] ->
    [ model_file := ref ]
END

(************************************************************************)

(** {6 Hint tactic } *)

let hint n =

  let open Proofview in
  Goal.nf_enter 
    begin fun gl ->

      let features_string = 
        Features.to_string ~sep:" " (collect_features (Goal.concl gl)) in

      (* Pp.msg_info (hov 0 (Printer.pr_type ())) ; *)

      let colt_prog = !colt_dir ^ "/colt" in
      let model_path = !colt_dir  ^ "/models/" ^ !model_file in
      let cmd = sprintf "%s suggest \"%s\" -m %s -n %i 2> log" 
        colt_prog features_string model_path n in

      if !debug_mode then
        Pp.msg_info (str cmd) ;

      (* if Sys.command cmd <> 0 then
        Pp.msg_info (str "Error in calling 'coq-learn'.") ; *)

      let ic = Unix.open_process_in cmd in
      let lines = read_lines ic in

      Pp.msg_info (prlist_with_sep Pp.fnl str lines) ;
      Pp.pp_flush ();
      Tacticals.New.tclIDTAC
    end


TACTIC EXTEND hint
  | [ "hint" ] -> [ hint 10 ]
  | [ "hint" integer(n) ] -> [ hint n ]
END


(************************************************************************)


(** {6 Additional stuff } *)


(* Export features using a custom program *)

let export_command = "add-features.sh"

let dump_dir = ref "features"

let export_features_custom gref = 
  let fs = features gref in
  let cmd = sprintf "%s '%s' '%s' \"%s\""
    (export_command)
    (!dump_dir)
    (string_of_gref gref)
    (Features.to_string fs) in
  if Sys.command cmd <> 0 then
    Pp.msgerrnl (str "Error.")

VERNAC COMMAND EXTEND SetFeaturesDir
  | ["Set" "Features" "Dir" string(s)] -> [ dump_dir := s ]
END


(* Print the body of a constant specified by a global reference. *)

let print_body gref = 
  let err () = Pp.msgerrnl (str "Not a constant.") in
  match gref with
  | Globnames.ConstRef cst ->
    begin
      let cb = Environ.lookup_constant cst (Global.env()) in
      match Global.body_of_constant_body cb with
      | Some e -> Pp.msgnl (Termops.print_constr e)
      | _ -> err ()
    end
  | _ -> err ()


VERNAC COMMAND EXTEND PrintBody
  | ["Print" "Body" reference(ref)] -> [ print_body (Nametab.global ref) ]
END

(************************************************************************)
